@@ -1,5 +1,3 @@
-//! # paxakos
-//!
 //! Paxakos is a pure Rust implementation of a distributed consensus algorithm
 //! based on Leslie Lamport's [Paxos][wikipedia]. It enables distributed systems
 //! to consistently modify shared state across their network, even in the
@@ -7,7 +5,18 @@
 //!
 //! [wikipedia]: https://en.wikipedia.org/wiki/Paxos_(computer_science)
 //!
-//! ## Usage
+//! [![crates.io][crates.io-badge]][crates.io-url]
+//! [![docs.rs][docs.rs-badge]][docs.rs-url]
+//! [![GPLv3 licensed][gpl-badge]][gpl-url]
+//!
+//! [crates.io-badge]: https://img.shields.io/crates/v/paxakos.svg
+//! [crates.io-url]: https://crates.io/crates/paxakos
+//! [docs.rs-badge]: https://docs.rs/paxakos/badge.svg
+//! [docs.rs-url]: https://docs.rs/paxakos
+//! [gpl-badge]: https://img.shields.io/badge/license-GPLv3-blue.svg
+//! [gpl-url]: LICENSE
+//!
+//! # Usage
 //!
 //! In order to use Paxakos, the traits [`LogEntry`], [`State`], [`NodeInfo`]
 //! and [`Communicator`][Communicator] need to be implemented. The first two
@@ -129,14 +138,14 @@
 //! }
 //! ```
 //!
-//! ## Motivation
+//! # Motivation
 //!
 //! Rust is a great language with which to implement efficient and truly
 //! reliable, fault-tolerant services. And while there already are several Rust
 //! implementations of consensus algorithms, they are either rudimentary or
 //! incomplete or their API was not approachable enough for this author.
 //!
-//! ### Priorities
+//! # Priorities
 //!
 //! The project's priorities are as follows.
 //!
@@ -169,19 +178,19 @@
 //!    correctness or simplicity. The biggest challenge in this area are, at
 //!    present, build times. If you have other concerns, please open an issue.
 //!
-//! ## Features
+//! # Features
 //!
 //! Paxakos is a Multi-Paxos implementation. It supports membership changes,
 //! concurrency, as well as taking and installing snapshots.
 //!
-//! ### Membership Changes
+//! ## Membership Changes
 //!
 //! The `State` trait exposes the [`cluster_at`][State::cluster_at] method. By
 //! implementing it, _arbitrary_ membership changes may be made. No restrictions
 //! are imposed and it is up to users and implementors to make sure that any
 //! transition is safe.
 //!
-//! ### Concurrency
+//! ## Concurrency
 //!
 //! Multi-Paxos allows for any number of rounds to be settled concurrently. This
 //! can be exploited by implementing `State`'s [concurrency][State::concurrency]
@@ -192,19 +201,7 @@
 //! This is typically done by appending no-op entries. A utility for doing so
 //! automatically is provided, but its API is far from stable.
 //!
-//! #### Interjection
-//!
-//! Consensus based clusters typically elect a single leader and who drive all
-//! progress. This is highly efficient, as each leader election incurs overhead.
-//! That notwithstanding, Paxakos has made the unusual design decision to allow
-//! multiple leaders at the same time, but for different rounds.
-//!
-//! This design allows a follower node to "interject" an entry at the end of the
-//! concurrency window. This part of the design hasn't been fleshed out yet, but
-//! it could allow nodes to conveniently queue operations without introducing
-//! additional communication protocols besides Paxakos.
-//!
-//! ### Snapshots
+//! ## Snapshots
 //!
 //! A node may take a snapshot of its current state. The result is a combination
 //! of the application specific `State` as well as pakakos specifc information.
@@ -212,7 +209,7 @@
 //! nodes which have joined the cluster or to catch up nodes that have fallen
 //! behind.
 //!
-//! ## Protocol
+//! # Protocol
 //!
 //! This section describes the Paxakos protocol. It is, for the most part, a
 //! typical _Multi-Paxos_ protocol. Multi-Paxos generalizes Paxos to be run for
@@ -244,13 +241,14 @@
 //!    stability and consistency and is achieved by exploiting the deterministic
 //!    order of nodes returned by [`cluster_at`][State::cluster_at].
 //!
-//! 2. Promise or Rejection
+//! 2. Vote
 //!
 //!    When a node receives a prepare request, it checks that it hasn't accepted
 //!    a previous such request with a coordination number that's higher than the
-//!    given one. If it hasn't, it sends back a _promise_ not to accept any more
-//!    _proposals_ with a coordination number that's less the given one.
-//!    Otherwise it sends back a rejection.
+//!    given one. If it has, it will reply with a conflict. If it hasn't, the
+//!    node will either abstain, i.e., choose not to give its vote, or it sends
+//!    back a _promise_ not to accept any more _proposals_ with a coordination
+//!    number that's less the given one.
 //!
 //!    1. Promise `(Vec<(RoundNum, CoordNum, LogEntry)>)`
 //!
@@ -259,12 +257,19 @@
 //!       acknowledge your bid to become leader and give you my vote. However,
 //!       in return you must respect these previous commitments I've made."
 //!
-//!    2. Rejection `(CoordNum, Option<(CoordNum, LogEntry)>)`
+//!    2. Conflict `(CoordNum, Option<(CoordNum, LogEntry)>)`
 //!
 //!       A rejection is sent with the highest observed coordination number so
 //!       far. For the special case that the round has already converged and the
 //!       node still has it available, it will send it along as well.
 //!
+//!    3. Abstention `A`
+//!
+//!       The node chose to abstain. By default nodes will never abstain. This
+//!       can be changed by providing a custom `Voter` implementation. The
+//!       argument type `A` is defined by `Communicator::Abstain` and
+//!       `Voter::Abstain`.
+//!
 //! 3. Propose `(RoundNum, CoordNum, LogEntry)`
 //!
 //!    When a node sent a `prepare(r, c)` request and received a quorum or more
@@ -278,56 +283,64 @@
 //!    the same round and coordination number will have the same log entry as
 //!    well.)
 //!
-//! 4. Acceptance or Rejection
+//! 4. Vote
 //!
 //!    When a node receives a proposal, it will check whether it has made any
-//!    conflicting promises. If it hasn't it will simply reply that it has
-//!    accepted the entry. Otherwise it will respond with a rejection.
+//!    conflicting promises. If it has, it responds with a conflict. If it
+//!    hasn't, it may choose to accept or reject the proposal and reply
+//!    accordingly.
 //!
-//!    1. Acceptance `()`
-//!    2. Rejection `(CoordNum, Option<(CoordNum, LogEntry)>)`
+//!    1. Acceptance `Y` / Rejection `N`
+//!
+//!       By default nodes accept all proposals with `Y = ()`. This can be
+//!       changed by providing a custom `Voter` implementation. The argument
+//!       types `Y` and `N` are defined by `Communicator::Yea`,
+//!       `Communicator::Nay`, `Voter::Yea` and `Voter::Nay`.
+//!
+//!    2. Conflict `(CoordNum, Option<(CoordNum, LogEntry)>)`
 //!
 //!       See 2.2.
 //!
-//! 5. Commit `(RoundNum, LogEntry::Id)` / CommitById `(RoundNum, LogEntry)`
+//! 5. Commit `(RoundNum, CoordNum, LogEntry::Id)` /
+//!    CommitById `(RoundNum, CoordNum, LogEntry)`
 //!
 //!    After having received a quorum of acceptances, the round has converged on
 //!    the proposed entry. The leader node will commit the entry locally and
 //!    inform other nodes as well. Nodes who sent an acceptance will only be
 //!    sent the log entry's id, others will receive the full entry.
 //!
-//! ## Project Status
+//! # Project Status
 //!
 //! This section examines different aspects of paxakos, how far along they are
 //! and what's missing or in need of improvement.
 //!
-//! ### ☀️ Consensus Implementation ☀️
+//! ## ☀️ Consensus Implementation ☀️
 //!
 //! The core algorithm of paxakos is well-tested and has been exercised a lot.
 //! There is reason to be confident in its correctness.
 //!
-//! ### ⛅ Passive Mode ⛅
+//! ## ⛅ Passive Mode ⛅
 //!
 //! [Passive mode][crate::node::Participation] is implemented and superficially
 //! tested. Thorough testing is still needed.
 //!
-//! ### 🌧️ Serialization 🌧️
+//! ## 🌧️ Serialization 🌧️
 //!
 //! Snapshot serialization needs to be reworked to allow for backward compat.
 //!
-//! ### ⛅ API Stability ⛅
+//! ## ⛅ API Stability ⛅
 //!
 //! The API has been fairly stable and there is no reason to expect big changes.
 //! The decorations may be reworked so that their configuration *can* be moved
 //! into the `State`.
 //!
-//! ### ☁️ Efficiency ☁️
+//! ## ⛅ Efficiency ⛅
 //!
-//! With its support for concurrency, paxakos has the potential to be very
-//! efficient. To fully unlock this potential, an implementation of master
-//! leases is required.
+//! Paxakos supports concurrency and has a (for now rudimentary) implementation
+//! of master leases. Assuming a scheme to delegate to the current master, this
+//! combination is highly efficient.
 //!
-//! ### Nightly Features
+//! ## Nightly Features
 //!
 //! Paxakos currently relies on several nightly features. None of them are
 //! crucial, but there hasn't been any need become compatible with stable Rust.
@@ -337,18 +350,13 @@
 //! This is a side project. I will work on the following as my time allows (in
 //! no particular order).
 //!
-//! - Passive Mode
-//!
-//! - Tests
-//!
-//! - Adding comments and documentation
-//!
-//! - Additional decorations
-//!
-//! - Support for _Master Leases_ (see section 5.2 of [Paxos Made
-//!   Live][paxos-made-live]).
-//!
-//! [paxos-made-live]: https://doi.org/10.1145/1281100.1281103
+//!  - Tests
+//!  - Adding comments and documentation
+//!  - Rounding off existing decorations
+//!  - Additional decorations, e.g.,
+//!    - for consistency checks
+//!    - for delegation to the current leader
+//!  - Serialization
 //!
 //! [Communicator]: crate::communicator::Communicator
 //! [LogEntry]: crate::log::LogEntry
@@ -361,6 +369,7 @@
 #![feature(array_windows)]
 #![feature(btree_drain_filter)]
 #![feature(drain_filter)]
+#![feature(duration_zero)]
 #![feature(map_first_last)]
 #![feature(never_type)]
 #![feature(step_trait)]
@@ -377,6 +386,11 @@ pub mod communicator;
 pub mod deco;
 pub mod error;
 pub mod event;
+pub mod invocation;
+#[cfg(feature = "master-leases")]
+pub mod leases;
+#[cfg(feature = "locking")]
+pub mod locking;
 mod log;
 pub mod node;
 #[cfg(feature = "prototyping")]
@@ -385,27 +399,38 @@ pub mod state;
 #[cfg(feature = "tracer")]
 pub mod tracer;
 mod util;
+pub mod voting;
 
-use std::convert::{TryFrom, TryInto};
+use std::convert::TryFrom;
+use std::convert::TryInto;
 use std::fmt::Debug;
 use std::hash::Hash;
 use std::sync::Arc;
 
-use serde::{Deserialize, Serialize};
+use serde::Deserialize;
+use serde::Serialize;
 
 // TODO move these three into the communicator module
+pub use error::AcceptError;
+pub use error::CommitError;
+pub use error::PrepareError;
+pub use event::Event;
+pub use event::ShutdownEvent;
 #[doc(inline)]
-pub use error::{AcceptError, CommitError, PrepareError};
-#[doc(inline)]
-pub use event::{Event, ShutdownEvent};
+pub use invocation::Invocation;
 #[doc(inline)]
 pub use log::LogEntry;
 #[doc(inline)]
 pub use node::builder as node_builder;
-#[doc(inline)]
-pub use node::{Commit, Node, NodeBuilder, NodeHandle, NodeInfo, NodeKernel, Shutdown};
-#[doc(inline)]
-pub use node::{NodeStatus, RequestHandler};
+pub use node::Commit;
+pub use node::Node;
+pub use node::NodeBuilder;
+pub use node::NodeHandle;
+pub use node::NodeInfo;
+pub use node::NodeKernel;
+pub use node::NodeStatus;
+pub use node::RequestHandler;
+pub use node::Shutdown;
 #[doc(inline)]
 pub use state::State;
 
@@ -474,18 +499,67 @@ pub trait Identifier: 'static + Copy + Debug + Eq + Hash + Send + Sync + Unpin {
 
 impl<T: 'static + Copy + Debug + Eq + Hash + Ord + Send + Sync + Unpin> Identifier for T {}
 
+/// A condition for a [Promise].
+#[derive(Clone, Debug, Serialize, Deserialize)]
+pub struct Condition<R, C, E> {
+    pub round_num: R,
+    pub coord_num: C,
+    pub log_entry: Arc<E>,
+}
+
+impl<R, C, E> From<Condition<R, C, E>> for (R, C, Arc<E>) {
+    fn from(val: Condition<R, C, E>) -> Self {
+        (val.round_num, val.coord_num, val.log_entry)
+    }
+}
+
+impl<R, C, E> From<(R, C, Arc<E>)> for Condition<R, C, E> {
+    fn from(condition: (R, C, Arc<E>)) -> Self {
+        let (round_num, coord_num, log_entry) = condition;
+
+        Self {
+            round_num,
+            coord_num,
+            log_entry,
+        }
+    }
+}
+
 /// A promise not to accept certain proposals anymore.
 ///
 /// Please refer to the [description of the protocol](crate#protocol).
 #[derive(Clone, Debug, Deserialize, Serialize)]
-pub struct Promise<R, C, E>(Vec<(R, C, Arc<E>)>);
+pub struct Promise<R, C, E>(Vec<Condition<R, C, E>>);
+
+impl<R: RoundNum, C, E> From<Vec<Condition<R, C, E>>> for Promise<R, C, E> {
+    fn from(mut conditions: Vec<Condition<R, C, E>>) -> Self {
+        conditions.sort_by_key(|c| c.round_num);
+
+        Self(conditions)
+    }
+}
+
+impl<R, C, E> From<Promise<R, C, E>> for Vec<Condition<R, C, E>> {
+    fn from(val: Promise<R, C, E>) -> Self {
+        val.0
+    }
+}
+
+impl<R, C, E> IntoIterator for Promise<R, C, E> {
+    type Item = Condition<R, C, E>;
+    type IntoIter = std::vec::IntoIter<Self::Item>;
+
+    fn into_iter(self) -> Self::IntoIter {
+        self.0.into_iter()
+    }
+}
+
+impl<R: RoundNum, C: CoordNum, E> Promise<R, C, E> {
+    /// The conditions this promise is predicated on.
+    pub fn conditions(&self) -> &[Condition<R, C, E>] {
+        &self.0
+    }
 
-impl<R, C, E> Promise<R, C, E>
-where
-    R: RoundNum,
-    C: CoordNum,
-    E: LogEntry,
-{
     pub(crate) fn empty() -> Self {
         Self(Vec::new())
     }
@@ -495,10 +569,11 @@ where
     }
 
     pub(crate) fn log_entry_for(&self, round_num: R) -> Option<Arc<E>> {
+        // TODO exploit order
         self.0
             .iter()
-            .find(|(r, _, _)| *r == round_num)
-            .map(|(_, _, e)| Arc::clone(e))
+            .find(|c| c.round_num == round_num)
+            .map(|c| Arc::clone(&c.log_entry))
     }
 
     pub(crate) fn merge_with(self, other: Self) -> Self {
@@ -510,40 +585,40 @@ where
         let mut a = a.into_iter();
         let mut b = b.into_iter();
 
-        let mut a_next = a.next();
-        let mut b_next = b.next();
+        let mut a_next = a.next().map(Into::into);
+        let mut b_next = b.next().map(Into::into);
 
         loop {
             match (a_next, b_next) {
                 (Some((ra, ca, ea)), Some((rb, cb, eb))) => match ra.cmp(&rb) {
                     std::cmp::Ordering::Equal => {
                         if ca > cb {
-                            max.push((ra, ca, ea));
+                            max.push(Condition::from((ra, ca, ea)));
                         } else {
-                            max.push((rb, cb, eb));
+                            max.push(Condition::from((rb, cb, eb)));
                         }
 
-                        a_next = a.next();
-                        b_next = b.next();
+                        a_next = a.next().map(Into::into);
+                        b_next = b.next().map(Into::into);
                     }
                     std::cmp::Ordering::Less => {
-                        max.push((ra, ca, ea));
-                        a_next = a.next();
+                        max.push(Condition::from((ra, ca, ea)));
+                        a_next = a.next().map(Into::into);
                         b_next = Some((rb, cb, eb));
                     }
                     std::cmp::Ordering::Greater => {
-                        max.push((rb, cb, eb));
-                        b_next = b.next();
+                        max.push(Condition::from((rb, cb, eb)));
+                        b_next = b.next().map(Into::into);
                         a_next = Some((ra, ca, ea));
                     }
                 },
                 (Some((ra, ca, ea)), None) => {
-                    max.push((ra, ca, ea));
+                    max.push(Condition::from((ra, ca, ea)));
                     max.extend(a);
                     return Promise(max);
                 }
                 (None, Some((rb, cb, eb))) => {
-                    max.push((rb, cb, eb));
+                    max.push(Condition::from((rb, cb, eb)));
                     max.extend(b);
                     return Promise(max);
                 }
@@ -555,12 +630,12 @@ where
     }
 }
 
-/// Rejection of a prepare request or a proposal.
+/// Conflict to a prepare request or a proposal.
 ///
 /// Please refer to the [description of the protocol](crate#protocol).
 #[derive(Clone, Debug, Deserialize, Serialize)]
-pub enum Rejection<C, E> {
-    Conflict {
+pub enum Conflict<C, E> {
+    Supplanted {
         coord_num: C,
     },
     Converged {
@@ -569,29 +644,11 @@ pub enum Rejection<C, E> {
     },
 }
 
-impl<C, E> Rejection<C, E>
-where
-    C: CoordNum,
-    E: LogEntry,
-{
+impl<C: CoordNum, E> Conflict<C, E> {
     pub(crate) fn coord_num(&self) -> C {
         match *self {
-            Rejection::Conflict { coord_num } => coord_num,
-            Rejection::Converged { coord_num, .. } => coord_num,
+            Conflict::Supplanted { coord_num } => coord_num,
+            Conflict::Converged { coord_num, .. } => coord_num,
         }
     }
 }
-
-impl<R, C, E> IntoIterator for Promise<R, C, E>
-where
-    R: RoundNum,
-    C: CoordNum,
-    E: LogEntry,
-{
-    type Item = (R, C, Arc<E>);
-    type IntoIter = std::vec::IntoIter<Self::Item>;
-
-    fn into_iter(self) -> Self::IntoIter {
-        self.0.into_iter()
-    }
-}
