@@ -1,69 +1,57 @@
-use std::future::Future;
+use futures::future::FutureExt;
+use futures::future::LocalBoxFuture;
+use futures::future::TryFutureExt;
 
-use futures::future::{FutureExt, LocalBoxFuture, TryFutureExt};
-
-use crate::communicator::{Communicator, CoordNumOf, RoundNumOf};
+use crate::communicator::Communicator;
 use crate::deco::Decoration;
 use crate::error::SpawnError;
+use crate::invocation::AbstainOf;
+use crate::invocation::CommunicationErrorOf;
+use crate::invocation::CoordNumOf;
+use crate::invocation::Invocation;
+use crate::invocation::LogEntryOf;
+use crate::invocation::NayOf;
+use crate::invocation::NodeIdOf;
+use crate::invocation::NodeOf;
+use crate::invocation::RoundNumOf;
+use crate::invocation::SnapshotFor;
+use crate::invocation::StateOf;
+use crate::invocation::YeaOf;
 use crate::log::LogKeeping;
-use crate::node::{self, Participation, RequestHandlerFor};
+use crate::node;
+use crate::node::Participation;
 #[cfg(feature = "tracer")]
-use crate::state::LogEntryIdOf;
-use crate::state::{ContextOf, NodeIdOf};
-#[cfg(feature = "tracer")]
-use crate::tracer::{Tracer, TracerFor};
-use crate::{Identifier, Node, State};
+use crate::tracer::Tracer;
+use crate::voting::IndiscriminateVoter;
+use crate::voting::IndiscriminateVoterFor;
+use crate::voting::Voter;
+use crate::Node;
+use crate::State;
 
-use super::snapshot::{Snapshot, SnapshotFor};
-use super::{CommunicatorOf, NodeKernel, StateOf};
-
-/// Builder to spawn a `Node`.
-///
-/// This API is badly desgined. Please have a look at and follow the
-/// documentation of [`node_builder()`](crate::node_builder()).
-// TODO This trait seems pointless, rename -WithAll to NodeBuilder?
-//      Overall this seems a bit messy with all the structs. Is there a
-//      better approach?
-pub trait NodeBuilder: Sized {
-    type Node: 'static + Node;
-    type Future: Future<Output = Result<(RequestHandlerFor<Self::Node>, Self::Node), SpawnError>>;
-
-    fn decorated_with<D>(self, arguments: <D as Decoration>::Arguments) -> NodeBuilderWithAll<D>
-    where
-        D: Decoration<
-            Decorated = Self::Node,
-            State = StateOf<Self::Node>,
-            Communicator = CommunicatorOf<Self::Node>,
-        >;
-
-    fn spawn(self) -> Self::Future
-    where
-        <Self::Node as Node>::State: State<Context = ()>,
-    {
-        self.spawn_in(())
-    }
-
-    fn spawn_in(self, context: ContextOf<StateOf<Self::Node>>) -> Self::Future;
-}
+use super::snapshot::Snapshot;
+use super::CommunicatorOf;
+use super::InvocationOf;
+use super::NodeKernel;
+use super::RequestHandlerFor;
 
 #[derive(Default)]
-pub struct NodeBuilderBlank;
+pub struct NodeBuilderBlank<I>(std::marker::PhantomData<I>);
 
-impl NodeBuilderBlank {
+impl<I: Invocation> NodeBuilderBlank<I> {
     pub fn new() -> Self {
-        Self
+        Self(std::marker::PhantomData)
     }
 
-    pub fn for_node<I: Identifier>(self, node_id: I) -> NodeBuilderWithNodeId<I> {
+    pub fn for_node(self, node_id: NodeIdOf<I>) -> NodeBuilderWithNodeId<I> {
         NodeBuilderWithNodeId { node_id }
     }
 }
 
-pub struct NodeBuilderWithNodeId<I: Identifier> {
-    node_id: I,
+pub struct NodeBuilderWithNodeId<I: Invocation> {
+    node_id: NodeIdOf<I>,
 }
 
-impl<I: Identifier> NodeBuilderWithNodeId<I> {
+impl<I: Invocation> NodeBuilderWithNodeId<I> {
     pub fn working_in(
         self,
         dir: impl AsRef<std::path::Path>,
@@ -82,19 +70,27 @@ impl<I: Identifier> NodeBuilderWithNodeId<I> {
     }
 }
 
-pub struct NodeBuilderWithNodeIdAndWorkingDir<I: Identifier> {
+pub struct NodeBuilderWithNodeIdAndWorkingDir<I: Invocation> {
     working_dir: Option<std::path::PathBuf>,
-    node_id: I,
+    node_id: NodeIdOf<I>,
 }
 
-impl<I: Identifier> NodeBuilderWithNodeIdAndWorkingDir<I> {
+impl<I: Invocation> NodeBuilderWithNodeIdAndWorkingDir<I> {
     pub fn communicating_via<C>(
         self,
         communicator: C,
-    ) -> NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<C>
+    ) -> NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<I, C>
     where
-        C: Communicator,
-        <C as Communicator>::Node: crate::NodeInfo<Id = I>,
+        C: Communicator<
+            Node = NodeOf<I>,
+            RoundNum = RoundNumOf<I>,
+            CoordNum = CoordNumOf<I>,
+            LogEntry = LogEntryOf<I>,
+            Error = CommunicationErrorOf<I>,
+            Yea = YeaOf<I>,
+            Nay = NayOf<I>,
+            Abstain = AbstainOf<I>,
+        >,
     {
         NodeBuilderWithNodeIdAndWorkingDirAndCommunicator {
             working_dir: self.working_dir,
@@ -104,31 +100,38 @@ impl<I: Identifier> NodeBuilderWithNodeIdAndWorkingDir<I> {
     }
 }
 
-pub struct NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<C: Communicator> {
+pub struct NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<I: Invocation, C: Communicator> {
     working_dir: Option<std::path::PathBuf>,
-    node_id: <<C as Communicator>::Node as crate::NodeInfo>::Id,
+    node_id: NodeIdOf<I>,
     communicator: C,
 }
 
-impl<C: Communicator> NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<C> {
-    /// Starts the node without any state.
-    pub fn without<S>(self) -> NodeBuilderWithAll<NodeKernel<S, C>>
-    where
-        S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    {
+impl<I, C> NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<I, C>
+where
+    I: Invocation,
+    C: Communicator<
+        Node = NodeOf<I>,
+        RoundNum = RoundNumOf<I>,
+        CoordNum = CoordNumOf<I>,
+        LogEntry = LogEntryOf<I>,
+        Error = CommunicationErrorOf<I>,
+        Yea = YeaOf<I>,
+        Nay = NayOf<I>,
+        Abstain = AbstainOf<I>,
+    >,
+{
+    /// Starts the node without any state and in passive mode.
+    pub fn without_state(self) -> NodeBuilder<NodeKernel<I, C>> {
         self.with_snapshot_and_participation(None, Participation::Passive)
     }
 
     /// Starts a new cluster with the given initial state.
     ///
     /// The round number will be [zero](num_traits::Zero).
-    pub fn with_initial_state<S>(
+    pub fn with_initial_state<S: Into<Option<StateOf<I>>>>(
         self,
-        initial_state: impl Into<Option<S>>,
-    ) -> NodeBuilderWithAll<NodeKernel<S, C>>
-    where
-        S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    {
+        initial_state: S,
+    ) -> NodeBuilder<NodeKernel<I, C>> {
         self.with_snapshot_and_participation(
             initial_state.into().map(Snapshot::initial),
             Participation::Active,
@@ -149,38 +152,29 @@ impl<C: Communicator> NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<C> {
     /// [`Last`]: crate::event::ShutdownEvent::Last
     /// [recovering_with]:
     /// NodeBuilderWithNodeIdAndWorkingDirAndCommunicator::recovering_with
-    pub fn resuming_from<S>(
+    pub fn resuming_from<S: Into<Option<SnapshotFor<I>>>>(
         self,
-        snapshot: impl Into<Option<Snapshot<S, RoundNumOf<C>, CoordNumOf<C>>>>,
-    ) -> NodeBuilderWithAll<NodeKernel<S, C>>
-    where
-        S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    {
-        self.with_snapshot_and_participation(snapshot.into(), Participation::Active)
+        snapshot: S,
+    ) -> NodeBuilder<NodeKernel<I, C>> {
+        self.with_snapshot_and_participation(snapshot, Participation::Active)
     }
 
     /// Resume operation from the given snapshot.
     ///
     /// The node will participate passively until it can be certain that it is
     /// not breaking any previous commitments.
-    pub fn recovering_with<S>(
+    pub fn recovering_with<S: Into<Option<SnapshotFor<I>>>>(
         self,
-        snapshot: impl Into<Option<Snapshot<S, RoundNumOf<C>, CoordNumOf<C>>>>,
-    ) -> NodeBuilderWithAll<NodeKernel<S, C>>
-    where
-        S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    {
-        self.with_snapshot_and_participation(snapshot.into(), Participation::Passive)
+        snapshot: S,
+    ) -> NodeBuilder<NodeKernel<I, C>> {
+        self.with_snapshot_and_participation(snapshot, Participation::Passive)
     }
 
     /// Resume operation without a snapshot.
     ///
     /// The node will participate passively until it can be certain that it is
     /// not breaking any previous commitments.
-    pub fn recovering_without<S>(self) -> NodeBuilderWithAll<NodeKernel<S, C>>
-    where
-        S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    {
+    pub fn recovering_without_state(self) -> NodeBuilder<NodeKernel<I, C>> {
         self.with_snapshot_and_participation(None, Participation::Passive)
     }
 
@@ -198,14 +192,11 @@ impl<C: Communicator> NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<C> {
     ///
     /// [recovering_with]:
     /// NodeBuilderWithNodeIdAndWorkingDirAndCommunicator::recovering_with
-    pub fn joining_with<S>(
+    pub fn joining_with<S: Into<Option<SnapshotFor<I>>>>(
         self,
-        snapshot: impl Into<Option<Snapshot<S, RoundNumOf<C>, CoordNumOf<C>>>>,
-    ) -> NodeBuilderWithAll<NodeKernel<S, C>>
-    where
-        S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    {
-        self.with_snapshot_and_participation(snapshot.into(), Participation::Active)
+        snapshot: S,
+    ) -> NodeBuilder<NodeKernel<I, C>> {
+        self.with_snapshot_and_participation(snapshot, Participation::Active)
     }
 
     /// Commence operation without a snapshot.
@@ -222,27 +213,29 @@ impl<C: Communicator> NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<C> {
     ///
     /// [recovering_without]:
     /// NodeBuilderWithNodeIdAndWorkingDirAndCommunicator::recovering_without
-    pub fn joining_without<S>(self) -> NodeBuilderWithAll<NodeKernel<S, C>>
-    where
-        S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    {
+    pub fn joining_without_state(self) -> NodeBuilder<NodeKernel<I, C>> {
         self.with_snapshot_and_participation(None, Participation::Active)
     }
 
-    fn with_snapshot_and_participation<S>(
+    #[doc(hidden)]
+    pub fn with_snapshot_and_participation<S: Into<Option<SnapshotFor<I>>>>(
         self,
-        snapshot: Option<Snapshot<S, RoundNumOf<C>, CoordNumOf<C>>>,
-        participation: Participation<RoundNumOf<C>>,
-    ) -> NodeBuilderWithAll<NodeKernel<S, C>>
-    where
-        S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    {
-        NodeBuilderWithAll {
+        snapshot: S,
+        participation: Participation<RoundNumOf<I>>,
+    ) -> NodeBuilder<NodeKernel<I, C>> {
+        if matches!(participation, Participation::PartiallyActive(_)) {
+            panic!("Unsupported: {:?}", participation);
+        }
+
+        let snapshot = snapshot.into();
+
+        NodeBuilder {
             working_dir: self.working_dir,
             node_id: self.node_id,
             communicator: self.communicator,
             snapshot,
             participation,
+            voter: IndiscriminateVoter::new(),
             log_keeping: Default::default(),
             finisher: Box::new(Ok),
 
@@ -252,25 +245,36 @@ impl<C: Communicator> NodeBuilderWithNodeIdAndWorkingDirAndCommunicator<C> {
     }
 }
 
-type Finisher<N> = dyn FnOnce(NodeKernel<StateOf<N>, CommunicatorOf<N>>) -> Result<N, SpawnError>;
+type Finisher<N> =
+    dyn FnOnce(NodeKernel<InvocationOf<N>, CommunicatorOf<N>>) -> Result<N, SpawnError>;
 
-pub struct NodeBuilderWithAll<N: Node> {
+pub struct NodeBuilder<N: Node, V = IndiscriminateVoterFor<N>> {
     working_dir: Option<std::path::PathBuf>,
-    node_id: NodeIdOf<StateOf<N>>,
+    node_id: node::NodeIdOf<N>,
+    voter: V,
     communicator: CommunicatorOf<N>,
-    snapshot: Option<SnapshotFor<N>>,
+    snapshot: Option<node::SnapshotFor<N>>,
     participation: Participation<node::RoundNumOf<N>>,
     log_keeping: LogKeeping,
     finisher: Box<Finisher<N>>,
 
     #[cfg(feature = "tracer")]
-    tracer: Option<Box<TracerFor<CommunicatorOf<N>>>>,
+    tracer: Option<Box<dyn Tracer<InvocationOf<N>>>>,
 }
 
-impl<S, C> NodeBuilderWithAll<NodeKernel<S, C>>
+impl<I, C, V> NodeBuilder<NodeKernel<I, C>, V>
 where
-    S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    C: Communicator,
+    I: Invocation,
+    C: Communicator<
+        Node = NodeOf<I>,
+        RoundNum = RoundNumOf<I>,
+        CoordNum = CoordNumOf<I>,
+        LogEntry = LogEntryOf<I>,
+        Error = CommunicationErrorOf<I>,
+        Yea = YeaOf<I>,
+        Nay = NayOf<I>,
+        Abstain = AbstainOf<I>,
+    >,
 {
     pub fn limiting_applied_entry_logs_to(mut self, limit: usize) -> Self {
         self.log_keeping.logs_kept = limit;
@@ -287,32 +291,58 @@ where
     }
 
     #[cfg(feature = "tracer")]
-    pub fn traced_by(
-        mut self,
-        tracer: impl Into<Box<dyn Tracer<RoundNumOf<C>, CoordNumOf<C>, LogEntryIdOf<S>>>>,
-    ) -> Self {
+    pub fn traced_by<T: Into<Box<dyn Tracer<I>>>>(mut self, tracer: T) -> Self {
         self.tracer = Some(tracer.into());
 
         self
     }
+
+    pub fn voting_with<T>(self, voter: T) -> NodeBuilder<NodeKernel<I, C>, T> {
+        NodeBuilder {
+            working_dir: self.working_dir,
+            node_id: self.node_id,
+            voter,
+            communicator: self.communicator,
+            snapshot: self.snapshot,
+            participation: self.participation,
+            log_keeping: self.log_keeping,
+            finisher: self.finisher,
+
+            #[cfg(feature = "tracer")]
+            tracer: self.tracer,
+        }
+    }
 }
 
-impl<N: 'static + Node> NodeBuilder for NodeBuilderWithAll<N> {
-    type Node = N;
-    type Future = LocalBoxFuture<'static, Result<(RequestHandlerFor<N>, N), SpawnError>>;
-
-    fn decorated_with<D>(self, arguments: <D as Decoration>::Arguments) -> NodeBuilderWithAll<D>
+impl<N, V> NodeBuilder<N, V>
+where
+    N: Node + 'static,
+    V: Voter<
+        State = node::StateOf<N>,
+        RoundNum = node::RoundNumOf<N>,
+        CoordNum = node::CoordNumOf<N>,
+        Abstain = node::AbstainOf<N>,
+        Yea = node::YeaOf<N>,
+        Nay = node::NayOf<N>,
+    >,
+{
+    pub fn decorated_with<D>(self, arguments: <D as Decoration>::Arguments) -> NodeBuilder<D, V>
     where
-        D: Decoration<Decorated = N, State = StateOf<N>, Communicator = CommunicatorOf<N>>,
+        D: Decoration<
+            Decorated = N,
+            Invocation = InvocationOf<N>,
+            Communicator = CommunicatorOf<N>,
+        >,
     {
         let finisher = self.finisher;
 
-        NodeBuilderWithAll {
+        NodeBuilder {
             working_dir: self.working_dir,
             node_id: self.node_id,
             communicator: self.communicator,
             snapshot: self.snapshot,
             participation: self.participation,
+            voter: self.voter,
             log_keeping: self.log_keeping,
             finisher: Box::new(move |x| ((finisher)(x)).and_then(|node| D::wrap(node, arguments))),
 
@@ -321,7 +351,17 @@ impl<N: 'static + Node> NodeBuilder for NodeBuilderWithAll<N> {
         }
     }
 
-    fn spawn_in(self, context: ContextOf<StateOf<N>>) -> Self::Future {
+    pub fn spawn(self) -> LocalBoxFuture<'static, Result<(RequestHandlerFor<N>, N), SpawnError>>
+    where
+        node::StateOf<N>: State<Context = ()>,
+    {
+        self.spawn_in(())
+    }
+
+    pub fn spawn_in(
+        self,
+        context: node::ContextOf<N>,
+    ) -> LocalBoxFuture<'static, Result<(RequestHandlerFor<N>, N), SpawnError>> {
         let finisher = self.finisher;
 
         NodeKernel::spawn(
@@ -330,6 +370,8 @@ impl<N: 'static + Node> NodeBuilder for NodeBuilderWithAll<N> {
             super::SpawnArgs {
                 context,
                 working_dir: self.working_dir,
+                node_id: self.node_id,
+                voter: self.voter,
                 snapshot: self.snapshot,
                 participation: self.participation,
                 log_keeping: self.log_keeping,
