@@ -1,5 +1,7 @@
-use std::cell::{Cell, RefCell};
-use std::collections::{HashMap, HashSet};
+use std::cell::Cell;
+use std::cell::RefCell;
+use std::collections::HashMap;
+use std::collections::HashSet;
 use std::convert::TryFrom;
 use std::future::Future;
 use std::ops::RangeInclusive;
@@ -10,48 +12,89 @@ use futures::future::FutureExt;
 use futures::lock::Mutex;
 use futures::stream::futures_unordered::FuturesUnordered;
 use futures::stream::StreamExt;
-use num_traits::{One, Zero};
+use num_traits::One;
+use num_traits::Zero;
 use tracing::debug;
 
-use crate::append::{AppendArgs, AppendError, Importance, Peeryness};
-use crate::applicable::{ApplicableTo, ProjectionOf};
-use crate::communicator::{AcceptanceOrRejection, Communicator, CoordNumOf, ErrorOf};
-use crate::communicator::{PromiseOrRejection, RoundNumOf};
-use crate::state::{LogEntryOf, NodeIdOf, NodeOf};
-use crate::{LogEntry, Promise, Rejection, State};
+use crate::append::AppendArgs;
+use crate::append::AppendError;
+use crate::append::Importance;
+use crate::append::Peeryness;
+use crate::applicable::ApplicableTo;
+use crate::communicator::Acceptance;
+use crate::communicator::Communicator;
+use crate::communicator::ErrorOf;
+use crate::communicator::Vote;
+use crate::invocation::AbstainOf;
+use crate::invocation::AcceptanceFor;
+use crate::invocation::CommitFor;
+use crate::invocation::CommunicationErrorOf;
+use crate::invocation::CoordNumOf;
+use crate::invocation::Invocation;
+use crate::invocation::LogEntryOf;
+use crate::invocation::NayOf;
+use crate::invocation::NodeIdOf;
+use crate::invocation::NodeOf;
+use crate::invocation::PromiseFor;
+use crate::invocation::RoundNumOf;
+use crate::invocation::StateOf;
+use crate::invocation::VoteFor;
+use crate::invocation::YeaOf;
+use crate::Conflict;
+use crate::LogEntry;
+use crate::Promise;
 
-use super::commits::{Commit, Commits};
+use super::commits::Commit;
+use super::commits::Commits;
 use super::state_keeper::StateKeeperHandle;
 use super::NodeInfo;
 
 struct ElectionToken;
 
-pub struct NodeInner<S, C>
+pub struct NodeInner<I, C>
 where
-    S: State,
-    C: Communicator,
+    I: Invocation,
+    C: Communicator<
+        Node = NodeOf<I>,
+        RoundNum = RoundNumOf<I>,
+        CoordNum = CoordNumOf<I>,
+        LogEntry = LogEntryOf<I>,
+        Error = CommunicationErrorOf<I>,
+        Yea = YeaOf<I>,
+        Nay = NayOf<I>,
+        Abstain = AbstainOf<I>,
+    >,
 {
-    id: NodeIdOf<S>,
+    id: NodeIdOf<I>,
     communicator: RefCell<C>,
 
-    state_keeper: StateKeeperHandle<S, RoundNumOf<C>, CoordNumOf<C>>,
+    state_keeper: StateKeeperHandle<I>,
 
     election_lock: Mutex<ElectionToken>,
-    term_start: Cell<RoundNumOf<C>>,
-    campaigned_on: Cell<CoordNumOf<C>>,
+    term_start: Cell<RoundNumOf<I>>,
+    campaigned_on: Cell<CoordNumOf<I>>,
 
     commits: Commits,
 }
 
-impl<S, C> NodeInner<S, C>
+impl<I, C> NodeInner<I, C>
 where
-    S: State<LogEntry = <C as Communicator>::LogEntry, Node = <C as Communicator>::Node>,
-    C: Communicator,
+    I: Invocation,
+    C: Communicator<
+        Node = NodeOf<I>,
+        RoundNum = RoundNumOf<I>,
+        CoordNum = CoordNumOf<I>,
+        LogEntry = LogEntryOf<I>,
+        Error = CommunicationErrorOf<I>,
+        Yea = YeaOf<I>,
+        Nay = NayOf<I>,
+        Abstain = AbstainOf<I>,
+    >,
 {
     pub fn new(
-        id: NodeIdOf<S>,
+        id: NodeIdOf<I>,
         communicator: C,
-        state_keeper: StateKeeperHandle<S, RoundNumOf<C>, CoordNumOf<C>>,
+        state_keeper: StateKeeperHandle<I>,
         commits: Commits,
     ) -> Self {
         Self {
@@ -68,15 +111,15 @@ where
         }
     }
 
-    pub fn id(&self) -> NodeIdOf<S> {
+    pub fn id(&self) -> NodeIdOf<I> {
         self.id
     }
 
-    pub async fn append<A: ApplicableTo<S>>(
+    pub async fn append<A: ApplicableTo<StateOf<I>>>(
         self: Rc<Self>,
         applicable: A,
-        args: AppendArgs<RoundNumOf<C>>,
-    ) -> Result<Commit<S, RoundNumOf<C>, ProjectionOf<A, S>>, AppendError> {
+        args: AppendArgs<I>,
+    ) -> Result<CommitFor<I, A>, AppendError<I>> {
         let log_entry = applicable.into_log_entry();
         let log_entry_id = log_entry.id();
 
@@ -100,9 +143,9 @@ where
 
     async fn append_actively(
         self: Rc<Self>,
-        log_entry: Arc<LogEntryOf<S>>,
-        mut args: AppendArgs<RoundNumOf<C>>,
-    ) -> Result<Commit<S, RoundNumOf<C>>, AppendError> {
+        log_entry: Arc<LogEntryOf<I>>,
+        mut args: AppendArgs<I>,
+    ) -> Result<CommitFor<I>, AppendError<I>> {
         let mut i: usize = 0;
 
         loop {
@@ -130,10 +173,10 @@ where
 
     async fn try_append(
         self: Rc<Self>,
-        log_entry: Arc<LogEntryOf<S>>,
-        round: RangeInclusive<RoundNumOf<C>>,
+        log_entry: Arc<LogEntryOf<I>>,
+        round: RangeInclusive<RoundNumOf<I>>,
         importance: Importance,
-    ) -> Result<Commit<S, RoundNumOf<C>>, AppendError> {
+    ) -> Result<CommitFor<I>, AppendError<I>> {
         let reservation = self.state_keeper.reserve_round_num(round).await?;
 
         let result = self
@@ -148,10 +191,10 @@ where
 
     async fn try_append_internal(
         self: Rc<Self>,
-        log_entry: Arc<LogEntryOf<S>>,
-        round_num: RoundNumOf<C>,
+        log_entry: Arc<LogEntryOf<I>>,
+        round_num: RoundNumOf<I>,
         importance: Importance,
-    ) -> Result<Commit<S, RoundNumOf<C>>, AppendError> {
+    ) -> Result<CommitFor<I>, AppendError<I>> {
         let node_id = self.id;
 
         let log_entry_id = log_entry.id();
@@ -167,7 +210,7 @@ where
         // The number of *additional* nodes that make a quorum.
         let quorum_prime = cluster.len() / 2;
 
-        let other_nodes: Vec<NodeOf<S>> = cluster
+        let other_nodes: Vec<NodeOf<I>> = cluster
             .into_iter()
             .filter(move |n| n.id() != node_id)
             .collect();
@@ -220,17 +263,17 @@ where
         &self,
         cluster_size: usize,
         own_node_idx: usize,
-    ) -> Result<CoordNumOf<C>, AppendError> {
+    ) -> Result<CoordNumOf<I>, AppendError<I>> {
         assert!(own_node_idx < cluster_size);
 
-        let cluster_size = CoordNumOf::<C>::try_from(cluster_size).unwrap_or_else(|_| {
+        let cluster_size = CoordNumOf::<I>::try_from(cluster_size).unwrap_or_else(|_| {
             panic!(
                 "Cannot convert cluster size `{}` into a coordination number.",
                 cluster_size
             )
         });
 
-        let own_node_ix = CoordNumOf::<C>::try_from(own_node_idx).unwrap_or_else(|_| {
+        let own_node_ix = CoordNumOf::<I>::try_from(own_node_idx).unwrap_or_else(|_| {
             panic!(
                 "Cannot convert `{}` into a coordination number.",
                 own_node_idx
@@ -255,12 +298,12 @@ where
 
     async fn ensure_leadership(
         &self,
-        round_num: RoundNumOf<C>,
-        coord_num: CoordNumOf<C>,
+        round_num: RoundNumOf<I>,
+        coord_num: CoordNumOf<I>,
         quorum_prime: usize,
-        other_nodes: &[NodeOf<S>],
+        other_nodes: &[NodeOf<I>],
         importance: Importance,
-    ) -> Result<Option<Arc<LogEntryOf<S>>>, AppendError> {
+    ) -> Result<Option<Arc<LogEntryOf<I>>>, AppendError<I>> {
         // This holds iff we already believe to be leader for `round_num`.
         if round_num >= self.term_start.get() && coord_num == self.campaigned_on.get() {
             return Ok(self.state_keeper.accepted_entry_of(round_num).await?);
@@ -318,12 +361,12 @@ where
                         .collect::<FuturesUnordered<_>>();
 
                     while let Some(response) = pending_responses.next().await {
-                        if let Ok(PromiseOrRejection::Rejection(rejection)) = response {
+                        if let Ok(Vote::Conflicted(rejection)) = response {
                             self.state_keeper
                                 .observe_coord_num(rejection.coord_num())
                                 .await?;
 
-                            if let Rejection::Converged {
+                            if let Conflict::Converged {
                                 log_entry: Some((coord_num, entry)),
                                 ..
                             } = &rejection
@@ -344,11 +387,11 @@ where
     async fn become_leader(
         &self,
         _election_token: &ElectionToken,
-        round_num: RoundNumOf<C>,
-        coord_num: CoordNumOf<C>,
+        round_num: RoundNumOf<I>,
+        coord_num: CoordNumOf<I>,
         quorum_prime: usize,
-        other_nodes: &[NodeOf<S>],
-    ) -> Result<Option<Arc<LogEntryOf<S>>>, AppendError> {
+        other_nodes: &[NodeOf<I>],
+    ) -> Result<Option<Arc<LogEntryOf<I>>>, AppendError<I>> {
         let own_promise = self
             .state_keeper
             .prepare_entry(round_num, coord_num)
@@ -362,7 +405,7 @@ where
             .map(|(_node, fut)| fut.map(|x| x));
 
         let promise_or_rejection = self
-            .await_promise_quorum_or_first_rejection(
+            .await_promise_quorum_or_first_conflict(
                 round_num,
                 quorum_prime,
                 pending_responses,
@@ -371,7 +414,7 @@ where
             .await;
 
         match promise_or_rejection? {
-            PromiseOrRejection::Promise(promise) => {
+            Vote::Given(promise) => {
                 let converged_log_entry = if promise.is_empty() {
                     None
                 } else {
@@ -383,7 +426,10 @@ where
                     self.state_keeper
                         .accept_entries(
                             coord_num,
-                            promise.into_iter().map(|(r, _, e)| (r, e)).collect(),
+                            promise
+                                .into_iter()
+                                .map(|c| (c.round_num, c.log_entry))
+                                .collect(),
                         )
                         .await?;
 
@@ -401,12 +447,12 @@ where
                 Ok(converged_log_entry)
             }
 
-            PromiseOrRejection::Rejection(rejection) => {
+            Vote::Conflicted(rejection) => {
                 self.state_keeper
                     .observe_coord_num(rejection.coord_num())
                     .await?;
 
-                if let Rejection::Converged {
+                if let Conflict::Converged {
                     log_entry: Some((coord_num, entry)),
                     ..
                 } = rejection
@@ -419,25 +465,24 @@ where
 
                 Err(AppendError::Lost)
             }
+
+            Vote::Abstained(_) => {
+                unreachable!()
+            }
         }
     }
 
-    async fn await_promise_quorum_or_first_rejection(
+    async fn await_promise_quorum_or_first_conflict(
         &self,
-        round_num: RoundNumOf<C>,
+        round_num: RoundNumOf<I>,
         quorum: usize,
         pending_responses: impl IntoIterator<
-            Item = impl Future<
-                Output = Result<
-                    PromiseOrRejection<RoundNumOf<C>, CoordNumOf<C>, LogEntryOf<S>>,
-                    ErrorOf<C>,
-                >,
-            >,
+            Item = impl Future<Output = Result<VoteFor<I>, ErrorOf<C>>>,
         >,
-        own_promise: Promise<RoundNumOf<C>, CoordNumOf<C>, LogEntryOf<S>>,
-    ) -> Result<PromiseOrRejection<RoundNumOf<C>, CoordNumOf<C>, LogEntryOf<S>>, AppendError> {
+        own_promise: PromiseFor<I>,
+    ) -> Result<VoteFor<I>, AppendError<I>> {
         if quorum == 0 {
-            return Ok(PromiseOrRejection::Promise(own_promise));
+            return Ok(Vote::Given(own_promise));
         }
 
         let mut pending_responses: FuturesUnordered<_> = pending_responses.into_iter().collect();
@@ -448,22 +493,35 @@ where
         let mut promises = 0;
         let mut max_promise = own_promise;
 
+        let mut abstentions = Vec::new();
+        let mut communication_errors = Vec::new();
+
         while let Some(response) = pending_responses.next().await {
             pending_len -= 1;
 
             match response {
-                Err(_) => {
+                Err(err) => {
+                    communication_errors.push(err);
+
                     if promises + pending_len < quorum {
-                        return Err(AppendError::NoQuorum);
+                        return Err(AppendError::NoQuorum {
+                            abstentions,
+                            communication_errors,
+                            rejections: Vec::new(),
+                        });
                     }
                 }
 
-                Ok(PromiseOrRejection::Rejection(rejection)) => {
+                Ok(Vote::Abstained(abstention)) => {
+                    abstentions.push(abstention);
+                }
+
+                Ok(Vote::Conflicted(rejection)) => {
                     self.state_keeper
                         .observe_coord_num(rejection.coord_num())
                         .await?;
 
-                    if let Rejection::Converged {
+                    if let Conflict::Converged {
                         log_entry: Some((coord_num, entry)),
                         ..
                     } = &rejection
@@ -473,15 +531,15 @@ where
                             .await?;
                     }
 
-                    return Ok(PromiseOrRejection::Rejection(rejection));
+                    return Ok(Vote::Conflicted(rejection));
                 }
 
-                Ok(PromiseOrRejection::Promise(promise)) => {
+                Ok(Vote::Given(promise)) => {
                     let current_max_promise = std::mem::replace(&mut max_promise, Promise::empty());
                     let new_max_promise = current_max_promise.merge_with(promise);
 
                     if promises + 1 >= quorum {
-                        return Ok(PromiseOrRejection::Promise(new_max_promise));
+                        return Ok(Vote::Given(new_max_promise));
                     } else {
                         promises += 1;
                         max_promise = new_max_promise;
@@ -495,25 +553,20 @@ where
 
     async fn propose_entry(
         &self,
-        round_num: RoundNumOf<C>,
-        coord_num: CoordNumOf<C>,
+        round_num: RoundNumOf<I>,
+        coord_num: CoordNumOf<I>,
         quorum_prime: usize,
-        other_nodes: &[NodeOf<S>],
-        log_entry: Arc<LogEntryOf<S>>,
+        other_nodes: &[NodeOf<I>],
+        log_entry: Arc<LogEntryOf<I>>,
     ) -> Result<
         (
-            HashSet<NodeIdOf<S>>,
-            HashSet<NodeIdOf<S>>,
+            HashSet<NodeIdOf<I>>,
+            HashSet<NodeIdOf<I>>,
             FuturesUnordered<
-                impl Future<
-                    Output = (
-                        NodeIdOf<S>,
-                        Result<AcceptanceOrRejection<CoordNumOf<C>, LogEntryOf<S>>, ErrorOf<C>>,
-                    ),
-                >,
+                impl Future<Output = (NodeIdOf<I>, Result<AcceptanceFor<I>, ErrorOf<C>>)>,
             >,
         ),
-        AppendError,
+        AppendError<I>,
     > {
         self.state_keeper
             .accept_entry(round_num, coord_num, Arc::clone(&log_entry))
@@ -525,7 +578,7 @@ where
             .send_proposal(other_nodes, round_num, coord_num, Arc::clone(&log_entry))
             .into_iter()
             .map(|(node, fut)| {
-                let node_id: NodeIdOf<S> = node.id();
+                let node_id: NodeIdOf<I> = node.id();
                 fut.map(move |r| (node_id, r))
             });
 
@@ -535,24 +588,19 @@ where
 
     async fn await_accepted_quorum<'a, P>(
         &self,
-        round_num: RoundNumOf<C>,
+        round_num: RoundNumOf<I>,
         quorum: usize,
         pending_responses: impl IntoIterator<Item = P>,
     ) -> Result<
         (
-            HashSet<NodeIdOf<S>>,
-            HashSet<NodeIdOf<S>>,
+            HashSet<NodeIdOf<I>>,
+            HashSet<NodeIdOf<I>>,
             FuturesUnordered<P>,
         ),
-        AppendError,
+        AppendError<I>,
     >
     where
-        P: Future<
-            Output = (
-                NodeIdOf<S>,
-                Result<AcceptanceOrRejection<CoordNumOf<C>, LogEntryOf<S>>, ErrorOf<C>>,
-            ),
-        >,
+        P: Future<Output = (NodeIdOf<I>, Result<AcceptanceFor<I>, ErrorOf<C>>)>,
     {
         if quorum == 0 {
             return Ok((
@@ -570,20 +618,23 @@ where
         let mut accepted = HashSet::new();
         let mut rejected_or_failed = HashSet::new();
 
+        let mut rejections = Vec::new();
+        let mut communication_errors = Vec::new();
+
         let mut conflict = None;
 
         while let Some((node_id, response)) = pending_responses.next().await {
             pending_len -= 1;
 
             match response {
-                Ok(AcceptanceOrRejection::Acceptance) => {
+                Ok(Acceptance::Given(_)) => {
                     accepted.insert(node_id);
 
                     if accepted.len() >= quorum {
                         return Ok((accepted, rejected_or_failed, pending_responses));
                     }
                 }
-                Ok(AcceptanceOrRejection::Rejection(Rejection::Converged {
+                Ok(Acceptance::Conflicted(Conflict::Converged {
                     log_entry: Some((coord_num, entry)),
                     ..
                 })) => {
@@ -594,12 +645,21 @@ where
                     return Err(AppendError::Converged);
                 }
                 rejection_or_failure => {
-                    if let Ok(AcceptanceOrRejection::Rejection(Rejection::Conflict { coord_num })) =
-                        rejection_or_failure
-                    {
-                        conflict = conflict
-                            .map(|c| std::cmp::max(c, coord_num))
-                            .or(Some(coord_num));
+                    match rejection_or_failure {
+                        Err(err) => {
+                            communication_errors.push(err);
+                        }
+                        Ok(Acceptance::Conflicted(Conflict::Supplanted { coord_num })) => {
+                            conflict = conflict
+                                .map(|c| std::cmp::max(c, coord_num))
+                                .or(Some(coord_num));
+                        }
+                        Ok(Acceptance::Refused(rejection)) => rejections.push(rejection),
+                        Ok(Acceptance::Given(_))
+                        | Ok(Acceptance::Conflicted(Conflict::Converged { .. })) => {
+                            // covered in outer match
+                            unreachable!()
+                        }
                     }
 
                     if accepted.len() + pending_len < quorum {
@@ -607,7 +667,11 @@ where
                             self.state_keeper.observe_coord_num(coord_num).await?;
                         }
 
-                        return Err(AppendError::NoQuorum);
+                        return Err(AppendError::NoQuorum {
+                            abstentions: Vec::new(),
+                            communication_errors,
+                            rejections,
+                        });
                     }
 
                     rejected_or_failed.insert(node_id);
@@ -621,22 +685,16 @@ where
     #[allow(clippy::too_many_arguments)]
     async fn commit_entry(
         self: Rc<Self>,
-        round_num: RoundNumOf<C>,
-        coord_num: CoordNumOf<C>,
-        log_entry: Arc<LogEntryOf<S>>,
-        mut other_nodes: Vec<NodeOf<S>>,
-        accepted: HashSet<NodeIdOf<S>>,
-        rejected_or_failed: HashSet<NodeIdOf<S>>,
+        round_num: RoundNumOf<I>,
+        coord_num: CoordNumOf<I>,
+        log_entry: Arc<LogEntryOf<I>>,
+        mut other_nodes: Vec<NodeOf<I>>,
+        accepted: HashSet<NodeIdOf<I>>,
+        rejected_or_failed: HashSet<NodeIdOf<I>>,
         mut pending_acceptances: FuturesUnordered<
-            impl 'static
-                + Future<
-                    Output = (
-                        NodeIdOf<S>,
-                        Result<AcceptanceOrRejection<CoordNumOf<C>, LogEntryOf<S>>, ErrorOf<C>>,
-                    ),
-                >,
+            impl 'static + Future<Output = (NodeIdOf<I>, Result<AcceptanceFor<I>, ErrorOf<C>>)>,
         >,
-    ) -> Result<Commit<S, RoundNumOf<C>>, AppendError> {
+    ) -> Result<CommitFor<I>, AppendError<I>> {
         let accepted = other_nodes
             .drain_filter(|n| accepted.contains(&n.id()))
             .collect::<Vec<_>>();
@@ -678,7 +736,7 @@ where
                 let node = pending_nodes_by_id.remove(&node_id).expect("pending node");
 
                 match response {
-                    Ok(AcceptanceOrRejection::Acceptance) => {
+                    Ok(Acceptance::Given(_)) => {
                         let commit = self
                             .communicator
                             .borrow_mut()
