@@ -1,44 +1,96 @@
 mod error;
 mod handle;
 mod io;
-mod macros;
 mod msg;
 mod working_dir;
 
 use std::collections::btree_map;
-use std::collections::{BTreeMap, HashMap, HashSet, VecDeque};
+use std::collections::BTreeMap;
+use std::collections::HashMap;
+use std::collections::HashSet;
+use std::collections::VecDeque;
+use std::convert::TryFrom;
 use std::fmt::Debug;
 use std::ops::RangeInclusive;
 use std::sync::Arc;
+use std::time::Instant;
 
-use futures::channel::{mpsc, oneshot};
+use futures::channel::mpsc;
+use futures::channel::oneshot;
 use futures::sink::SinkExt;
 use futures::stream::StreamExt;
-use num_traits::{Bounded, One, Zero};
+use num_traits::Bounded;
+use num_traits::One;
+use num_traits::Zero;
 use pin_project::pin_project;
-use tracing::{debug, info};
+use smallvec::SmallVec;
+use tracing::debug;
+use tracing::info;
 
-use crate::error::{AcceptError, AffirmSnapshotError, CommitError, InstallSnapshotError};
-use crate::error::{PrepareError, PrepareSnapshotError, ReadStaleError, SpawnError};
-use crate::event::{Event, ShutdownEvent};
+use crate::error::AcceptError;
+use crate::error::AffirmSnapshotError;
+use crate::error::CommitError;
+use crate::error::InstallSnapshotError;
+use crate::error::PrepareError;
+use crate::error::PrepareSnapshotError;
+use crate::error::ReadStaleError;
+use crate::error::SpawnError;
+use crate::event::Event;
+use crate::event::ShutdownEvent;
+use crate::invocation::AbstainOf;
+use crate::invocation::ContextOf;
+use crate::invocation::CoordNumOf;
+use crate::invocation::Invocation;
+use crate::invocation::LogEntryIdOf;
+use crate::invocation::LogEntryOf;
+use crate::invocation::NayOf;
+use crate::invocation::NodeIdOf;
+use crate::invocation::NodeOf;
+use crate::invocation::OutcomeOf;
+use crate::invocation::PromiseFor;
+use crate::invocation::RoundNumOf;
+use crate::invocation::SnapshotFor;
+use crate::invocation::StateOf;
+use crate::invocation::YeaOf;
 use crate::log::LogEntry;
-use crate::state::{ContextOf, LogEntryIdOf, LogEntryOf, OutcomeOf, State};
+use crate::node::NodeInfo;
+use crate::state::State;
 #[cfg(feature = "tracer")]
 use crate::tracer::Tracer;
-use crate::{CoordNum, Promise, RoundNum};
+use crate::voting;
+use crate::voting::Voter;
+use crate::Condition;
+use crate::CoordNum;
+use crate::Promise;
+use crate::RoundNum;
 
-use super::snapshot::{DeconstructedSnapshot, Snapshot};
+use super::snapshot::DeconstructedSnapshot;
+use super::snapshot::Snapshot;
 use super::status::NodeStatus;
 use super::SpawnArgs;
 
-use error::{AcquireRoundNumError, ClusterError};
+use error::AcquireRoundNumError;
+use error::ClusterError;
 pub use handle::StateKeeperHandle;
-use msg::{Release, Request, Response};
+use msg::Release;
+use msg::Request;
+use msg::Response;
 use working_dir::WorkingDir;
 
-type RequestAndResponseSender<S, R, C> = (Request<S, R, C>, ResponseSender<S, R, C>);
-type ResponseSender<S, R, C> = oneshot::Sender<Response<S, R, C>>;
-type Awaiter<S, R> = oneshot::Sender<(R, OutcomeOf<S>)>;
+type RequestAndResponseSender<I> = (Request<I>, ResponseSender<I>);
+type ResponseSender<I> = oneshot::Sender<Response<I>>;
+type Awaiter<I> = oneshot::Sender<(RoundNumOf<I>, OutcomeOf<I>)>;
+
+// this should become unnecessary with stabilization of `!`
+macro_rules! assert_unreachable {
+    ($v:expr) => {{
+        h($v);
+
+        fn h(_: std::convert::Infallible) -> ! {
+            unreachable!()
+        }
+    }};
+}
 
 #[derive(Debug)]
 enum Participation<R, C> {
@@ -65,18 +117,41 @@ impl<R, C> From<super::Participation<R>> for Participation<R, C> {
     }
 }
 
-#[derive(Debug)]
-pub struct StateKeeper<S: State, R: RoundNum, C: CoordNum> {
-    context: ContextOf<S>,
-    working_dir: Option<WorkingDir<R>>,
+enum AcceptPolicy<I: Invocation> {
+    Irrejectable,
+    Rejectable(Option<NodeOf<I>>),
+}
 
-    receiver: mpsc::Receiver<RequestAndResponseSender<S, R, C>>,
+type SpawnResult<I> = Result<(NodeStatus, super::Participation<RoundNumOf<I>>), SpawnError>;
+type RoundNumRequest<I> = (RangeInclusive<RoundNumOf<I>>, ResponseSender<I>);
+type AcceptedEntry<I> = (CoordNumOf<I>, Arc<LogEntryOf<I>>);
+type PendingCommit<I> = (Instant, CoordNumOf<I>, Arc<LogEntryOf<I>>);
 
-    release_sender: mpsc::UnboundedSender<Release<R>>,
-    release_receiver: mpsc::UnboundedReceiver<Release<R>>,
+pub struct StateKeeper<I, V>
+where
+    I: Invocation,
+    V: Voter<
+        State = StateOf<I>,
+        RoundNum = RoundNumOf<I>,
+        CoordNum = CoordNumOf<I>,
+        Yea = YeaOf<I>,
+        Nay = NayOf<I>,
+        Abstain = AbstainOf<I>,
+    >,
+{
+    context: ContextOf<I>,
+    working_dir: Option<WorkingDir<RoundNumOf<I>>>,
+
+    node_id: NodeIdOf<I>,
+    voter: V,
+
+    receiver: mpsc::Receiver<RequestAndResponseSender<I>>,
+
+    release_sender: mpsc::UnboundedSender<Release<RoundNumOf<I>>>,
+    release_receiver: mpsc::UnboundedReceiver<Release<RoundNumOf<I>>>,
 
     // TODO persist in snapshots
-    highest_observed_round_num: Option<R>,
+    highest_observed_round_num: Option<RoundNumOf<I>>,
 
     /// The highest coordination _observed_ number so far.
     ///
@@ -86,7 +161,7 @@ pub struct StateKeeper<S: State, R: RoundNum, C: CoordNum> {
     /// bid for leader. This coordination number will then be compared against
     /// the _strongest_ promise (see prepare_entry). Therefore it makes no sense
     /// to track observations by round number.
-    highest_observed_coord_num: C,
+    highest_observed_coord_num: CoordNumOf<I>,
 
     /// If promises were made, promises will be kept.
     ///
@@ -116,41 +191,56 @@ pub struct StateKeeper<S: State, R: RoundNum, C: CoordNum> {
     /// enforcing the invariant above.
     ///
     /// [concurrency]: crate::state::State::concurrency
-    promises: BTreeMap<R, C>,
-    accepted_entries: BTreeMap<R, (C, Arc<LogEntryOf<S>>)>,
-    leadership: (R, C),
+    promises: BTreeMap<RoundNumOf<I>, CoordNumOf<I>>,
+    accepted_entries: BTreeMap<RoundNumOf<I>, AcceptedEntry<I>>,
+    leadership: (RoundNumOf<I>, CoordNumOf<I>),
 
     /// Round number of the last applied log entry, initially `Zero::zero()`.
-    state_round: R,
+    state_round: RoundNumOf<I>,
     /// Current state or `None` iff the node is `Disoriented`.
-    state: Option<Arc<S>>,
+    state: Option<Arc<StateOf<I>>>,
 
-    round_num_requests: VecDeque<(RangeInclusive<R>, ResponseSender<S, R, C>)>,
-    acquired_round_nums: HashSet<R>,
+    round_num_requests: VecDeque<RoundNumRequest<I>>,
+    acquired_round_nums: HashSet<RoundNumOf<I>>,
 
-    pending_commits: BTreeMap<R, (std::time::Instant, C, Arc<LogEntryOf<S>>)>,
-    awaiters: HashMap<LogEntryIdOf<S>, Vec<Awaiter<S, R>>>,
+    pending_commits: BTreeMap<RoundNumOf<I>, PendingCommit<I>>,
+    awaiters: HashMap<LogEntryIdOf<I>, Vec<Awaiter<I>>>,
 
-    participation: Participation<R, C>,
+    // TODO persist in snapshots
+    participation: Participation<RoundNumOf<I>, CoordNumOf<I>>,
     /// last status that was observable from the outside
     status: NodeStatus,
-    event_emitter: mpsc::Sender<ShutdownEvent<S, R, C>>,
 
-    applied_entry_buffer: VecDeque<(C, Arc<LogEntryOf<S>>)>,
+    queued_events: VecDeque<ShutdownEvent<I>>,
+    event_emitter: mpsc::Sender<ShutdownEvent<I>>,
+    warn_counter: u16,
+
+    applied_entry_buffer: VecDeque<(CoordNumOf<I>, Arc<LogEntryOf<I>>)>,
 
     #[cfg(feature = "tracer")]
-    tracer: Option<Box<dyn Tracer<R, C, LogEntryIdOf<S>>>>,
+    tracer: Option<Box<dyn Tracer<I>>>,
 }
 
-impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
+impl<I, V> StateKeeper<I, V>
+where
+    I: Invocation,
+    V: Voter<
+        State = StateOf<I>,
+        RoundNum = RoundNumOf<I>,
+        CoordNum = CoordNumOf<I>,
+        Yea = YeaOf<I>,
+        Nay = NayOf<I>,
+        Abstain = AbstainOf<I>,
+    >,
+{
     pub async fn spawn(
-        args: SpawnArgs<S, R, C>,
+        args: SpawnArgs<I, V>,
     ) -> Result<
         (
             NodeStatus,
-            super::Participation<R>,
-            EventStream<S, R, C>,
-            StateKeeperHandle<S, R, C>,
+            super::Participation<RoundNumOf<I>>,
+            EventStream<I>,
+            StateKeeperHandle<I>,
             ProofOfLife,
         ),
         SpawnError,
@@ -179,16 +269,16 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
     }
 
     fn start_and_run_new(
-        spawn_args: SpawnArgs<S, R, C>,
-        start_result_sender: oneshot::Sender<
-            Result<(NodeStatus, super::Participation<R>), SpawnError>,
-        >,
-        receiver: mpsc::Receiver<RequestAndResponseSender<S, R, C>>,
-        event_emitter: mpsc::Sender<ShutdownEvent<S, R, C>>,
+        spawn_args: SpawnArgs<I, V>,
+        start_result_sender: oneshot::Sender<SpawnResult<I>>,
+        receiver: mpsc::Receiver<RequestAndResponseSender<I>>,
+        event_emitter: mpsc::Sender<ShutdownEvent<I>>,
     ) {
         std::thread::spawn(move || {
             let context = spawn_args.context;
             let working_dir = spawn_args.working_dir;
+            let node_id = spawn_args.node_id;
+            let voter = spawn_args.voter;
             let snapshot = spawn_args.snapshot;
             let participation = spawn_args.participation;
             let log_keeping = spawn_args.log_keeping;
@@ -255,6 +345,9 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                 context,
                 working_dir,
 
+                node_id,
+                voter,
+
                 receiver,
 
                 release_sender: rel_send,
@@ -278,7 +371,10 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
 
                 participation: participation.into(),
                 status: initial_status,
+
+                queued_events: VecDeque::new(),
                 event_emitter,
+                warn_counter: 0,
 
                 // TODO the capacity should be configurable
                 applied_entry_buffer: VecDeque::with_capacity(1024),
@@ -292,14 +388,11 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
     }
 
     fn init_and_run(mut self) {
-        crate::emit!(
-            self,
-            ShutdownEvent::Regular(Event::Init {
-                status: self.status,
-                round: self.state_round,
-                state: self.state.as_ref().map(Arc::clone),
-            })
-        );
+        self.emit(Event::Init {
+            status: self.status,
+            round: self.state_round,
+            state: self.state.as_ref().map(Arc::clone),
+        });
 
         self.run();
     }
@@ -309,29 +402,15 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
             self.apply_commits();
             self.release_round_nums();
             self.hand_out_round_nums();
+            self.detect_and_emit_status_change();
 
-            let expected_status = match &self.state {
-                Some(state) => {
-                    if self
-                        .highest_observed_round_num
-                        .unwrap_or_else(Bounded::max_value)
-                        > self.state_round + into_round_num(state.concurrency())
-                    {
-                        NodeStatus::Lagging
-                    } else if self.highest_observed_coord_num == self.leadership.1 {
-                        NodeStatus::Leading
-                    } else {
-                        NodeStatus::Following
-                    }
-                }
-                None => NodeStatus::Disoriented,
-            };
-
-            if self.status != expected_status {
-                self.emit_status_change(expected_status);
+            if !self.flush_event_queue() {
+                tracing::warn!("Node was not properly shut down but simply dropped.");
+                return;
             }
         }
 
+        tracing::info!("Shutting down.");
         self.shut_down()
     }
 
@@ -427,6 +506,53 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
         }
     }
 
+    fn flush_event_queue(&mut self) -> bool {
+        while let Some(event) = self.queued_events.pop_front() {
+            let mut timeout = 1;
+
+            let mut send = self.event_emitter.send(event);
+
+            loop {
+                timeout *= 10;
+
+                let timeout = std::time::Duration::from_millis(timeout);
+                let delay = futures_timer::Delay::new(timeout);
+
+                match futures::executor::block_on(futures::future::select(send, delay)) {
+                    futures::future::Either::Left((Ok(_), _)) => break,
+                    futures::future::Either::Left((Err(_), _)) => return false,
+
+                    futures::future::Either::Right((_, s)) => {
+                        self.warn_counter += 1;
+
+                        const LIMIT: u16 = 10;
+
+                        if self.warn_counter <= LIMIT
+                            || std::time::Duration::from_secs(60) < timeout
+                        {
+                            tracing::warn!(
+                                "`{}` events have been queued for over `{:?}`.",
+                                self.queued_events.len() + 1,
+                                timeout
+                            );
+                        }
+
+                        if self.warn_counter == LIMIT {
+                            tracing::warn!(
+                                "Limit of `{}` warnings reached, further warnings will be suppressed.",
+                                LIMIT
+                            );
+                        }
+
+                        send = s;
+                    }
+                }
+            }
+        }
+
+        true
+    }
+
     fn shut_down(mut self) {
         let snapshot = self.prepare_snapshot().ok();
         let mut emitter = self.event_emitter;
@@ -434,11 +560,7 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
         let _ = futures::executor::block_on(emitter.send(ShutdownEvent::Last { snapshot }));
     }
 
-    fn handle_request_msg(
-        &mut self,
-        req: Request<S, R, C>,
-        resp_sender: oneshot::Sender<Response<S, R, C>>,
-    ) {
+    fn handle_request_msg(&mut self, req: Request<I>, resp_sender: oneshot::Sender<Response<I>>) {
         let resp = match req {
             Request::PrepareSnapshot => Response::PrepareSnapshot(self.prepare_snapshot()),
 
@@ -551,7 +673,12 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                     .map(|(r, _)| *r)
                     .find(|r| *r > self.state_round);
 
-                let result = self.accept_entries(coord_num, entries);
+                let leader = self.deduce_node(entries[0].0, coord_num);
+                if let Some(leader) = &leader {
+                    assert_eq!(leader.id(), self.node_id);
+                }
+
+                let result = self.accept_entries(coord_num, entries, AcceptPolicy::Irrejectable);
 
                 if let Some(round_num) = round_num {
                     if result.is_ok() {
@@ -559,7 +686,7 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                     }
                 }
 
-                Response::AcceptEntries(result)
+                Response::AcceptEntries(result.map(|_| ()))
             }
 
             Request::CommitEntry {
@@ -635,7 +762,7 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
         self.participation = Participation::passive();
     }
 
-    fn passive_for(&self, round_num: R) -> bool {
+    fn passive_for(&self, round_num: RoundNumOf<I>) -> bool {
         match self.participation {
             Participation::Active => false,
             Participation::PartiallyActive(r) => round_num < r,
@@ -643,21 +770,82 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
         }
     }
 
+    fn detect_and_emit_status_change(&mut self) {
+        let expected_status = match &self.state {
+            Some(state) => {
+                if self
+                    .highest_observed_round_num
+                    .unwrap_or_else(Bounded::max_value)
+                    > self.state_round + into_round_num(state.concurrency())
+                {
+                    NodeStatus::Lagging
+                } else if self.highest_observed_coord_num == self.leadership.1 {
+                    NodeStatus::Leading
+                } else {
+                    NodeStatus::Following
+                }
+            }
+            None => NodeStatus::Disoriented,
+        };
+
+        if self.status != expected_status {
+            self.emit_status_change(expected_status);
+        }
+    }
+
     fn emit_status_change(&mut self, new_status: NodeStatus) {
         let old_status = self.status;
         self.status = new_status;
 
-        crate::emit!(
-            self,
-            ShutdownEvent::Regular(Event::StatusChange {
-                old_status,
-                new_status
-            })
-        );
+        self.emit(Event::StatusChange {
+            old_status,
+            new_status,
+        });
     }
 
-    fn emit_directive(&mut self, round_num: R, coord_num: C) {
-        if let Some(state) = &self.state {
+    fn emit_directive(&mut self, round_num: RoundNumOf<I>, coord_num: CoordNumOf<I>) {
+        if let Some(leader) = self.deduce_node(round_num, coord_num) {
+            self.emit(Event::Directive {
+                leader,
+                round_num,
+                coord_num,
+                timestamp: Instant::now(),
+            });
+        }
+    }
+
+    fn emit_gaps(&mut self, mut last_round: RoundNumOf<I>) {
+        let mut gaps = Vec::new();
+
+        for (&k, v) in self.pending_commits.iter() {
+            if k > last_round + One::one() {
+                gaps.push(crate::event::Gap {
+                    since: v.0,
+                    rounds: (last_round + One::one())..k,
+                });
+            }
+
+            last_round = k;
+        }
+
+        self.emit(Event::Gaps(gaps));
+    }
+
+    fn emit(&mut self, event: impl Into<ShutdownEvent<I>> + std::fmt::Debug) {
+        tracing::trace!("Emitting event {:?}.", event);
+
+        let event = event.into();
+
+        match self.event_emitter.try_send(event) {
+            Ok(_) => {}
+            Err(err) => {
+                self.queued_events.push_back(err.into_inner());
+            }
+        }
+    }
+
+    fn deduce_node(&self, round_num: RoundNumOf<I>, coord_num: CoordNumOf<I>) -> Option<NodeOf<I>> {
+        self.state.as_ref().and_then(|state| {
             if round_num > self.state_round
                 && round_num <= self.state_round + into_round_num(state.concurrency())
             {
@@ -667,7 +855,7 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                     .expect("Zero was ruled out via equality check");
 
                 let cluster = state.cluster_at(round_offset);
-                let cluster_size = C::try_from(cluster.len()).unwrap_or_else(|_| {
+                let cluster_size = CoordNumOf::<I>::try_from(cluster.len()).unwrap_or_else(|_| {
                     panic!(
                         "Cannot convert cluster size `{}` into a coordination number.",
                         cluster.len()
@@ -678,32 +866,27 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                         panic!("Out of usize range: {}", coord_num % cluster_size);
                     });
 
-                let leader = cluster[leader_ix].clone();
+                let leader = cluster.into_iter().nth(leader_ix).unwrap();
 
-                crate::emit!(
-                    self,
-                    ShutdownEvent::Regular(Event::Directive {
-                        leader,
-                        round_num,
-                        coord_num
-                    })
-                );
+                Some(leader)
+            } else {
+                None
             }
-        }
+        })
     }
 
-    fn observe_round_num(&mut self, round_num: R) {
+    fn observe_round_num(&mut self, round_num: RoundNumOf<I>) {
         self.highest_observed_round_num = self
             .highest_observed_round_num
             .map(|r| std::cmp::max(r, round_num))
             .or(Some(round_num));
     }
 
-    fn observe_coord_num(&mut self, coord_num: C) {
+    fn observe_coord_num(&mut self, coord_num: CoordNumOf<I>) {
         self.highest_observed_coord_num = std::cmp::max(self.highest_observed_coord_num, coord_num);
     }
 
-    fn prepare_snapshot(&mut self) -> Result<Snapshot<S, R, C>, PrepareSnapshotError> {
+    fn prepare_snapshot(&mut self) -> Result<SnapshotFor<I>, PrepareSnapshotError> {
         let state = self
             .state
             .as_ref()
@@ -719,14 +902,11 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
     }
 
     // TODO this doesn't do anything and probably never will, consider removal
-    fn affirm_snapshot(&mut self, _snapshot: Snapshot<S, R, C>) -> Result<(), AffirmSnapshotError> {
+    fn affirm_snapshot(&mut self, _snapshot: SnapshotFor<I>) -> Result<(), AffirmSnapshotError> {
         Ok(())
     }
 
-    fn install_snapshot(
-        &mut self,
-        snapshot: Snapshot<S, R, C>,
-    ) -> Result<(), InstallSnapshotError> {
+    fn install_snapshot(&mut self, snapshot: SnapshotFor<I>) -> Result<(), InstallSnapshotError> {
         if self.state_round > snapshot.round() {
             return Err(InstallSnapshotError::Outdated);
         }
@@ -747,13 +927,10 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
         self.promises = promises;
         self.accepted_entries = accepted_entries;
 
-        crate::emit!(
-            self,
-            ShutdownEvent::Regular(Event::Install {
-                round: state_round,
-                state
-            })
-        );
+        self.emit(Event::Install {
+            round: state_round,
+            state,
+        });
 
         let extraneous_commits: Vec<_> = self
             .pending_commits
@@ -769,9 +946,9 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
 
     fn prepare_entry(
         &mut self,
-        round_num: R,
-        coord_num: C,
-    ) -> Result<Promise<R, C, LogEntryOf<S>>, PrepareError<S, C>> {
+        round_num: RoundNumOf<I>,
+        coord_num: CoordNumOf<I>,
+    ) -> Result<PromiseFor<I>, PrepareError<I>> {
         if self.passive_for(round_num) {
             debug!("In passive mode, rejecting prepare request.");
             return Err(PrepareError::Passive);
@@ -849,32 +1026,51 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
         // constraint here, these attempts would fail. That would prevent the node from
         // making progress and thereby break liveness.
         if coord_num < strongest_promise {
-            Err(PrepareError::Conflict(strongest_promise))
+            Err(PrepareError::Supplanted(strongest_promise))
         } else {
-            overriding_insert(&mut self.promises, round_num, coord_num);
+            let voting_decision = self.voter.contemplate_candidate(
+                round_num,
+                coord_num,
+                self.deduce_node(round_num, coord_num).as_ref(),
+                self.state.as_deref(),
+            );
 
-            let promise = self
-                .accepted_entries
-                .range(round_num..)
-                .map(|(r, (c, e))| (*r, *c, Arc::clone(e)))
-                .collect::<Vec<_>>();
+            match voting_decision {
+                voting::Decision::Abstain(reason) => Err(PrepareError::Abstained(reason)),
+                voting::Decision::Nay(never) => assert_unreachable!(never),
+                voting::Decision::Yea(_) => {
+                    overriding_insert(&mut self.promises, round_num, coord_num);
 
-            #[cfg(feature = "tracer")]
-            {
-                if let Some(tracer) = self.tracer.as_mut() {
-                    tracer.record_promise(
-                        round_num,
-                        coord_num,
-                        promise.iter().map(|(r, c, e)| (*r, *c, e.id())).collect(),
-                    );
+                    let promise = self
+                        .accepted_entries
+                        .range(round_num..)
+                        .map(|(r, (c, e))| Condition::from((*r, *c, Arc::clone(e))))
+                        .collect::<Vec<Condition<_, _, _>>>();
+
+                    #[cfg(feature = "tracer")]
+                    {
+                        if let Some(tracer) = self.tracer.as_mut() {
+                            tracer.record_promise(
+                                round_num,
+                                coord_num,
+                                promise
+                                    .iter()
+                                    .map(|c| (c.round_num, c.coord_num, c.log_entry.id()))
+                                    .collect(),
+                            );
+                        }
+                    }
+
+                    Ok(Promise(promise))
                 }
             }
-
-            Ok(Promise(promise))
         }
     }
 
-    fn try_get_applied_entry(&mut self, round_num: R) -> Option<(C, Arc<LogEntryOf<S>>)> {
+    fn try_get_applied_entry(
+        &mut self,
+        round_num: RoundNumOf<I>,
+    ) -> Option<(CoordNumOf<I>, Arc<LogEntryOf<I>>)> {
         if let Some(offset) = crate::util::try_usize_delta(self.state_round, round_num) {
             if let Some((c, e)) = self.applied_entry_buffer.get(offset) {
                 return Some((*c, Arc::clone(e)));
@@ -890,10 +1086,10 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
 
     fn accept_entry(
         &mut self,
-        round_num: R,
-        coord_num: C,
-        entry: Arc<LogEntryOf<S>>,
-    ) -> Result<(), AcceptError<S, C>> {
+        round_num: RoundNumOf<I>,
+        coord_num: CoordNumOf<I>,
+        entry: Arc<LogEntryOf<I>>,
+    ) -> Result<YeaOf<I>, AcceptError<I>> {
         if round_num <= self.state_round {
             Err(AcceptError::Converged(
                 self.highest_observed_coord_num,
@@ -912,19 +1108,28 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                 overridable_lookup(&self.promises, round_num, "relevant promise");
 
             if relevant_promise > coord_num {
-                Err(AcceptError::Conflict(relevant_promise))
+                Err(AcceptError::Supplanted(relevant_promise))
             } else {
-                self.accept_entries(coord_num, vec![(round_num, entry)])
-                    .map(|n| assert!(n == 1))
+                self.accept_entries(
+                    coord_num,
+                    vec![(round_num, entry)],
+                    AcceptPolicy::Rejectable(self.deduce_node(round_num, coord_num)),
+                )
+                .map(|ys| {
+                    assert!(ys.len() == 1);
+
+                    ys.into_iter().next().unwrap()
+                })
             }
         }
     }
 
     fn accept_entries(
         &mut self,
-        coord_num: C,
-        entries: Vec<(R, Arc<LogEntryOf<S>>)>,
-    ) -> Result<usize, AcceptError<S, C>> {
+        coord_num: CoordNumOf<I>,
+        entries: Vec<(RoundNumOf<I>, Arc<LogEntryOf<I>>)>,
+        policy: AcceptPolicy<I>,
+    ) -> Result<SmallVec<[YeaOf<I>; 1]>, AcceptError<I>> {
         let first_round = entries[0].0;
 
         if let Participation::Passive {
@@ -932,7 +1137,6 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
         } = &mut self.participation
         {
             for r in entries.iter().map(|(r, _)| *r) {
-                println!("Adding {:?}", (r, coord_num));
                 observed_proposals.insert((r, coord_num));
             }
         }
@@ -942,8 +1146,6 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
             return Err(AcceptError::Passive);
         }
 
-        let mut accepted = 0;
-
         let round_range_start = self.state_round + One::one();
         let round_range_end = self
             .promises
@@ -955,12 +1157,33 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
 
         let round_range = round_range_start
             ..=(round_range_end
+                // TODO is the max here necessary? maybe add an assertion of r >= 1
                 .map(|r| std::cmp::max(r, One::one()) - One::one())
                 .unwrap_or_else(Bounded::max_value));
 
         let acceptable_entries = entries.into_iter().filter(|(r, _)| round_range.contains(r));
 
+        let mut yeas = SmallVec::new();
+
         for (round_num, log_entry) in acceptable_entries {
+            if let AcceptPolicy::Rejectable(leader) = &policy {
+                match self.voter.contemplate_proposal(
+                    round_num,
+                    coord_num,
+                    &*log_entry,
+                    leader.as_ref(),
+                    self.state.as_deref(),
+                ) {
+                    voting::Decision::Abstain(never) => assert_unreachable!(never),
+                    voting::Decision::Nay(rejection) => {
+                        return Err(AcceptError::Rejected(rejection))
+                    }
+                    voting::Decision::Yea(yea) => {
+                        yeas.push(yea);
+                    }
+                }
+            }
+
             #[cfg(feature = "tracer")]
             {
                 if let Some(tracer) = self.tracer.as_mut() {
@@ -968,8 +1191,6 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                 }
             }
 
-            accepted += 1;
-
             match self.accepted_entries.entry(round_num) {
                 btree_map::Entry::Occupied(ref mut e) => {
                     // At this point we've accepted the entry. However, when a node makes a promise
@@ -986,15 +1207,15 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
             }
         }
 
-        Ok(accepted)
+        Ok(yeas)
     }
 
     fn commit_entry(
         &mut self,
-        round_num: R,
-        coord_num: C,
-        entry: Arc<LogEntryOf<S>>,
-    ) -> Result<(), CommitError<S>> {
+        round_num: RoundNumOf<I>,
+        coord_num: CoordNumOf<I>,
+        entry: Arc<LogEntryOf<I>>,
+    ) -> Result<(), CommitError<I>> {
         #[cfg(feature = "tracer")]
         {
             if let Some(tracer) = self.tracer.as_mut() {
@@ -1011,21 +1232,25 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
             return Err(CommitError::Disoriented);
         }
 
-        crate::emit!(
-            self,
-            ShutdownEvent::Regular(Event::Commit {
-                round: round_num,
-                log_entry: Arc::clone(&entry),
-            })
+        self.voter.observe_commit(
+            round_num,
+            coord_num,
+            &entry,
+            self.deduce_node(round_num, coord_num).as_ref(),
         );
 
+        self.emit(Event::Commit {
+            round: round_num,
+            log_entry: Arc::clone(&entry),
+        });
+
         // Make sure that a shrinking gap doesn't get younger.
         let gap_time = self
             .pending_commits
             .range(round_num..)
             .next()
             .map(|e| e.1 .0)
-            .unwrap_or_else(std::time::Instant::now);
+            .unwrap_or_else(Instant::now);
 
         debug!("Queuing entry {:?} for round {}.", entry, round_num);
         self.pending_commits
@@ -1033,7 +1258,7 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
 
         // Are we missing commits between this one and the last one applied?
         if round_num > self.state_round + One::one() {
-            crate::emit_gaps!(self, self.state_round);
+            self.emit_gaps(self.state_round);
         }
 
         Ok(())
@@ -1051,7 +1276,7 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
         self.state = Some(state);
     }
 
-    fn apply_commits_to(&mut self, state: Arc<S>) -> (R, Arc<S>) {
+    fn apply_commits_to(&mut self, state: Arc<StateOf<I>>) -> (RoundNumOf<I>, Arc<StateOf<I>>) {
         if !self
             .pending_commits
             .contains_key(&(self.state_round + One::one()))
@@ -1084,16 +1309,11 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                 // round `r + c`, regardless of any potential concurrency increases between now
                 // (`r`) and then (`r + c`) because no node could have seen them before (`r` was
                 // not settled).
-                println!("Checking for {:?}", (round, coord_num));
                 if observed_proposals.contains(&(round, coord_num)) {
                     let first_active_round = round + into_round_num(state.concurrency());
                     self.participation = Participation::PartiallyActive(first_active_round);
 
-                    println!("{:?}", Event::<S, R, C>::Activate(first_active_round));
-                    crate::emit!(
-                        self,
-                        ShutdownEvent::Regular(Event::Activate(first_active_round))
-                    );
+                    self.emit(Event::Activate(first_active_round));
                 }
             }
 
@@ -1128,20 +1348,17 @@ impl<S: State, R: RoundNum, C: CoordNum> StateKeeper<S, R, C> {
                 working_dir.log_entry_application(round, coord_num, &*entry);
             }
 
-            crate::emit!(
-                self,
-                ShutdownEvent::Regular(Event::Apply {
-                    round,
-                    log_entry: entry,
-                    result: event
-                })
-            );
+            self.emit(Event::Apply {
+                round,
+                log_entry: entry,
+                result: event,
+            });
         }
 
         self.clean_up_after_applies();
 
         if !self.pending_commits.is_empty() {
-            crate::emit_gaps!(self, round);
+            self.emit_gaps(round);
         }
 
         (round, Arc::new(state))
@@ -1224,13 +1441,13 @@ impl ProofOfLife {
 }
 
 #[pin_project]
-pub struct EventStream<S: State, R: RoundNum, C: CoordNum> {
+pub struct EventStream<I: Invocation> {
     #[pin]
-    delegate: mpsc::Receiver<ShutdownEvent<S, R, C>>,
+    delegate: mpsc::Receiver<ShutdownEvent<I>>,
 }
 
-impl<S: State, R: RoundNum, C: CoordNum> futures::stream::Stream for EventStream<S, R, C> {
-    type Item = ShutdownEvent<S, R, C>;
+impl<I: Invocation> futures::stream::Stream for EventStream<I> {
+    type Item = ShutdownEvent<I>;
 
     fn poll_next(
         self: std::pin::Pin<&mut Self>,
